/*
   DIO.C - Direct Input/Output routines.

*/
#include <stdio.h>
#include <stdlib.h>

#include "dio.h"

#if (PLATFORM == _BORLAND_C) || (PLATFORM == _DJGPP_C)
   #include <conio.h>
   #include <dos.h>
#elif (PLATFORM == _AIX_C)
   #include <curses.h>
#endif

/* headers needed only by DJGPP */
#if (PLATFORM == _DJGPP_C)
   #include <pc.h>   /* used for sound() and nosound() */
#endif



/*
   Global variables.
*/
char
   __pad_character__,  /* what character do I pad dio_display() with */
   __sound__,          /* Make noise when input buffer full dio_edit()? */
   __backspace_exit__; /* does backspacing in dio_edit() exit function? */

int
   __default_exit_keys__[] = { ENTER, TAB, UP, DOWN, PGUP, PGDN, F1, F2,
      F3, F4, F5, F6, F7, F8, F9, F10},
   __exit_keys__[50];


/*
   Initializes the direct I/O.
*/
void dio_on(void)
{
   dio_padchar(' ');
   dio_textcolor(-1);
   dio_sound(1);
   dio_backspace_exit(0); /* backspace won't exit dio_edit() */
   dio_setup_exit_keys(__default_exit_keys__, 16);
#if (PLATFORM == _AIX_C)
   initscr();
   noecho();
   cbreak();
   keypad(stdscr, 1);
#endif
}


/*
   Shuts down the direct I/O.
*/
void dio_off(void)
{
#if (PLATFORM == _BORLAND_C) || (PLATFORM == _DJGPP_C)
   clrscr();
   /* restore colors to default */
   dio_textcolor(-1);
   dio_textbackground(-1);

#elif (PLATFORM == _AIX_C)
   refresh();
   endwin();
#endif
}


/*
   Sets up the exit keys for dio_edit().
*/
void dio_setup_exit_keys(int *keys, int num_keys)
{
   int
      i;

   for(i = 0; i < num_keys; i++)
      __exit_keys__[i] = keys[i];
   __exit_keys__[i] = 0;
}


/*
   Returns the number of rows on the display device.
*/
int dio_rows(void)
{
#if (PLATFORM == _BORLAND_C) || (PLATFORM == _DJGPP_C)
   struct text_info
      screen_info;

   gettextinfo(&screen_info);

   return screen_info.screenheight;
#elif (PLATFORM == _AIX_C)
   return LINES;
#endif
}


/*
   Returns the number of columns on the display device.
*/
int dio_cols(void)
{
#if (PLATFORM == _BORLAND_C) || (PLATFORM == _DJGPP_C)
   struct text_info
      screen_info;

   gettextinfo(&screen_info);

   return screen_info.screenwidth;
#elif (PLATFORM == _AIX_C)
   return COLS;
#endif
}


/*
   Clears the screen and puts cursor in upper right hand corner.
*/
void dio_clear(void)
{
#if (PLATFORM == _BORLAND_C) || (PLATFORM == _DJGPP_C)
   clrscr();
#elif (PLATFORM == _AIX_C)
   erase();
#endif
}


/*
   Emties the screen output buffer.

   Note:  Borland version does nothing.
*/
void dio_update(void)
{
#if (PLATFORM == _AIX_C)
   refresh();
#endif
}


/*
   Updates the screen and waits for a key press.
*/
int dio_getchar(void)
{
   /* must put variable definitions before function calls. */
#if (PLATFORM == _BORLAND_C) || (PLATFORM == _DJGPP_C)
   int
      key;
#endif

   dio_update();

#if (PLATFORM == _BORLAND_C) || (PLATFORM == _DJGPP_C)
   key = getch();

   /* if key is extended then return scan code + 1000 */
   key = (key == 0) ? (getch() + 1000) : key;

   return key;
#elif (PLATFORM == _AIX_C)
   return getch();
#endif
}


/*
   Moves the cursor to (row, column).

   Note: (0, 0) is upper left hand corner of screen.
*/
void dio_moveto(int row, int column)
{
#if (PLATFORM == _BORLAND_C) || (PLATFORM == _DJGPP_C)
   /* Borland uses (1, 1) for upper left hand corner of screen */
   gotoxy(column + 1, row + 1);
#elif (PLATFORM == _AIX_C)
   move(row, column);
#endif
}


/*
   Puts a character on screen at the current cursor position and
   cursor is moved one position to the right.
*/
void dio_putchar(int c)
{
#if (PLATFORM == _BORLAND_C) || (PLATFORM == _DJGPP_C)
   putch(c);
#elif (PLATFORM == _AIX_C)
   addch(c);
#endif
}


/*
   Puts a string on the screen starting at the current cursor postion.
   The cursor is then moved to the position just after the string end.
*/
void dio_puts(const char *s)
{
#if (PLATFORM == _BORLAND_C) || (PLATFORM == _DJGPP_C)
   cputs(s);
#elif (PLATFORM == _AIX_C)
   addstr(s);
#endif
}


/*
   Calculates the length of a string and returns it.
*/
int __string_length__(const char *s)
{
   register int
      cnt;

   for(cnt = 0; s[cnt++];) { }

   /* must subtract 1 from cnt or it return length of string + NULL */
   return --cnt;
}


/*
   Displays a string at starting at (row, column) for a maximum length
   of (length).  If the string is longer that (length) then only
   (length) amount of characters are shown.  If the string is shorter
   then (length) characters then trailing spaces are added.
*/
void dio_display(const char *s, int row, int column, int length)
{
   int
      spaces_to_pad,
      str_pos,
      str_length,
      num_chars_to_print;


   dio_moveto(row, column);

   str_length = __string_length__(s);

   /* calculate number of characters to print and spaces to pad. */
   num_chars_to_print = (str_length >= length) ? length : str_length;
   spaces_to_pad = length - num_chars_to_print;

   /* Print string */
   for(str_pos = 0; str_pos < num_chars_to_print; str_pos++)
   {
      dio_putchar(s[str_pos]);
   }

   /* Pad left over length (if any) with spaces */
   for(str_pos = 0; str_pos < spaces_to_pad; str_pos++)
   {
      dio_putchar(__pad_character__);
   }
}

/*
   Allocates the memory and copies a string in memory.
   The address of the copy is returned.  On failure the
   function returns NULL;
*/
char *__make_string_copy__(const char *source)
{
   char
      *dest;
   int
      len;

   len = __string_length__(source);
   dest = (char *)malloc(len + 1);  /* +1 for NULL */
   if(dest)
   {
      memcpy(dest, source, len);
   }

   return dest;
}


/*
   Displays the string to edit on the screen.
*/
void __display_string__(char *s, int string_offset, int row,
                        int col, int flen)
{
   dio_display((s + string_offset), row, col, flen);
}


/*
   Deletes a character from the string.  All characters to the left of
   (position_in_string) are shifted to the left 1 place.
*/
void __delete_character__(char *s, int position_in_string)
{
   int
      end,
      str_pos;

   if(s[position_in_string] != 0)  /* if I'm not at the NULL character */
   {
      end = __string_length__(s) - 1;

      for(str_pos = position_in_string; str_pos <= end; str_pos++)
      {
         s[str_pos] = s[str_pos + 1];  /* shift characters one to the left */
      }
   }
}


/*
   Toggles the insert mode.

   Note:  This was put in function for future enhancement such as
          displaying an indication of the insertion mode or changing
          the cursor size and or shape.
*/
void __insert_mode__(int *mode)
{
   *mode ^= 1;

   /* Make cursor change shape */
#if (PLATFORM == _BORLAND_C) || (PLATFORM == _DJGPP_C)
   _setcursortype(((*mode == 1) ? _NORMALCURSOR : _SOLIDCURSOR));
#endif
}


/*
   Moves the cursor the begining of the string.
*/
void __home_key__(int *string_offset, int *field_offset, int row, int col)
{
   *string_offset = 0;
   *field_offset = 0;

   dio_moveto(row, col);
}


/*
   Moves the cursor to the end of the string.
*/
void __end_key__(int *string_offset, int *field_offset, int flen,
                 int row, int col, int str_len)
{
   if(str_len >= flen)
   {
      /* Make sure that the cursor is put one position past the end
         of the string, but still in the field. */
      *string_offset = str_len - (flen - 1);
      *field_offset = flen - 1;
   } else
   {
      *string_offset = 0;
      *field_offset = str_len;
   }

   dio_moveto(row, *field_offset + col);
}


/*
   Moves the cursor to the left one position if not at the begining of the
   string.
*/
void __left_key__(int *string_offset, int *field_offset, int row, int col)
{
   if((*string_offset + *field_offset) != 0)  /* if not at begining */
   {
      if(*field_offset == 0)
      {
         *string_offset -= 1;
      } else
      {
         *field_offset -= 1;
      }

      dio_moveto(row, *field_offset + col);
   }
}


/*
   Removes a character from the string one position before the cursor.
*/
void __backspace__(char *s, int *string_offset, int *field_offset,
                   int row, int col)
{
   if((*string_offset + *field_offset) != 0) /* if I'm not at the begining */
   {
      __delete_character__(s, (*string_offset + *field_offset - 1));
      __left_key__(string_offset, field_offset, row, col);
   }
}


/*
   Moves the cursor to the right one position if not at the end of the
   string.
*/
void __right_key__(int *string_offset, int *field_offset, int flen,
                 int row, int col, int str_len)
{
   /* if not at end of string */
   if((*string_offset + *field_offset) != str_len)
   {
      /* If already at end of field. Scroll string */
      if(*field_offset == (flen - 1))
      {
         *string_offset += 1;
      } else  /* just adjust position in field. */
      {
         *field_offset += 1;
      }

      dio_moveto(row, *field_offset + col);
   }
}


/*
   Inserts a character into the string and moves the characters
   from (position) to the end of the string one position to the left.
*/
void __insert_character__(char *s, int position, int c)
{
   int
      end,
      str_pos;

   end = __string_length__(s) + 1;

   for(str_pos = end; str_pos >= position; str_pos--)
   {
      s[str_pos + 1] = s[str_pos];
   }

   s[position] = (char)c;
}

/*
   Makes a beep sound.

   Note:  Only works for DJGPP.
*/
void __beep__(void)
{
#if (PLATFORM == _DJGPP_C || PLATFORM == _BORLAND_C)
   if(__sound__)   /* play warning sound if it is turned on */
   {
      sound(300);
      delay(125);
      nosound();
   }
#endif
}

/*
   Adds a character to the string based on the insertion mode.
*/
void __add_character__(char *s, int *string_offset, int *field_offset,
                       int flen, int slen, int row, int col,
                       int insert_mode, int c)
{
   int
      position_in_string,
      string_length;

   position_in_string = *string_offset + *field_offset;
   string_length = strlen(s);

   if(!s[position_in_string]) /* at end of string */
   {
      if(string_length < slen) /* space still left in string to add */
      {
         __insert_character__(s, position_in_string, c);
         __right_key__(string_offset, field_offset, flen,
                       row, col, (string_length + 1));
      } else   /* No more room in string */
      {
         __beep__();
      }
   } else   /* some where in middle of string */
   {
      if(insert_mode)
      {
         if(string_length < slen) /* still room to add to string */
         {
            __insert_character__(s, position_in_string, c);
            __right_key__(string_offset, field_offset, flen,
                          row, col, (string_length + 1));
         } else   /* No more room in string */
         {
            __beep__();
         }
      } else   /* overstrike mode */
      {
         s[position_in_string] = (char)c;
         __right_key__(string_offset, field_offset, flen,
                       row, col, string_length);
      }
   }
}



/*
    Returns 1 if the (key) was one that is an exit key.
*/
int __valid_exit_key__(int key)
{
   register int
      cnt;
   int
      valid_key_to_exit = 0;

   /* check until end of list or match found. */
   for(cnt = 0; (__exit_keys__[cnt] && !valid_key_to_exit); cnt++)
   {
      valid_key_to_exit = (key == __exit_keys__[cnt]);
   }

   return valid_key_to_exit;
}

/*
   Performs full string editing on the string (s) at the location (row, col).
   If ESCAPE is pressed then editing is aborted.

   This function will only return DELETE as a key if delete was pressed on
   a 0-length string.  The same applies to backspace.
*/
int dio_edit(char *s, int row, int col, int flen, int slen)
{
   char
      *abort_copy;
   int
      string_offset,   /* position in string that appears at (col)    */
      field_offset,    /* number of places past (col) */
      insert_on = 1,   /* is the insert on? */
      key,
      done = 0;        /* done editing yet? */


   string_offset = 0;
   field_offset = 0;

   abort_copy = __make_string_copy__(s);

   do
   {
      __display_string__(s, string_offset, row, col, flen);
      dio_moveto(row, field_offset + col);
      dio_update();

      key = dio_getchar();

      switch(key)
      {
         case DELETE:
         {
            /* s[string_offset + field_offset] is checked so that
               if the position i'm at is not NULL then I cannot have
               a zero length string.  This saves a call to
               __string_length__ */
            if(__backspace_exit__)
            {
               if((s[string_offset + field_offset] == 0) &&  /* if this string */
                  (__string_length__(s) == 0))               /* blank, return */
               {
                  done = 1;
               } else
               {
                  __delete_character__(s, (string_offset + field_offset));
               }
            } else
            {
               __delete_character__(s, (string_offset + field_offset));
            }
            break;
         }
         case INSERT:
         {
            __insert_mode__(&insert_on);
            break;
         }
         case HOME:
         {
            __home_key__(&string_offset, &field_offset, row, col);
            break;
         }
         case END:
         {
            __end_key__(&string_offset, &field_offset, flen, row,
                        col, __string_length__(s));
            break;
         }
         case RIGHT:
         {
            __right_key__(&string_offset, &field_offset, flen, row,
                          col, __string_length__(s));
            break;
         }
         case LEFT:
         {
            __left_key__(&string_offset, &field_offset, row, col);
            break;
         }
         case BACKSPACE:
         {
            /* s[string_offset + field_offset] is checked so that
               if the position i'm at is not NULL then I cannot have
               a zero length string.  This saves a call to
               __string_length__ */

            if(__backspace_exit__)
            {
               if((s[string_offset + field_offset] == 0) && /* if string */
                  (__string_length__(s) == 0))              /* blank return */
               {
                  done = 1;
               } else
               {
                  __backspace__(s, &string_offset, &field_offset, row, col);
               }
            } else
            {
               __backspace__(s, &string_offset, &field_offset, row, col);
            }
            break;
         }
         case ESCAPE:
         {
            /* Restore the original string if enough memory was available
               to save it. */
            if(abort_copy)
            {
               memcpy(s, abort_copy, slen);
            }
            done = 1;
            break;
         }
         default:
         {
            if(__valid_exit_key__(key))  /* if it was an exit key end loop */
            {
               done = 1;
            } else
            {
               /* if printable then try to add it to the string */
               if(PRINTABLE_CHARACTER(key))
               {
                  __add_character__(s, &string_offset, &field_offset,
                                    flen, slen, row, col, insert_on, key);
               }
            }
         }
      }
   } while(!done);

   free(abort_copy);

   return key;
}


/*
   Sets the pad character that is used by dio_display().
*/
void dio_padchar(char c)
{
   __pad_character__ = c;
}


/*
   Sets the text foreground color.

   Using -1 will reset the color to the default.

   Note: on the AIX this function does nothing
*/
void dio_textcolor(int color)
{
#if (PLATFORM == _BORLAND_C) || (PLATFORM == _DJGPP_C)
   textcolor(((color == -1) ? LIGHTGRAY : color));
#endif
}


/*
   Sets the text background color.

   Using -1 will reset the color to the default.

   Note: on the AIX this function does nothing
*/
void dio_textbackground(int color)
{
#if (PLATFORM == _BORLAND_C) || (PLATFORM == _DJGPP_C)
   textbackground(((color == -1) ? BLACK : color));
#endif
}


/*
    Toggles the warning sound when dio_edit() buffer is full.

    Note: For AIX this function does nothing.
*/
void dio_sound(int state)
{
#if (PLATFORM == _DJGPP_C || PLATFORM == _BORLAND_C)
   __sound__ = state;
#endif
}

/*
    Sets the backspace for (or not) being able to exit dio_edit() when the
    string length is 0.
*/
void dio_backspace_exit(int state)
{
   __backspace_exit__ = state;
}
